""" Package logging functions and constants. """

import argparse
import logging
import os
import sys
from ._version import __version__


__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


__all__ = ["add_logging_options", "logger_via_cli", "setup_logger",
           "DEV_LOGGING_FMT", "DEVMODE_OPTNAME",
           "TRACE_LEVEL_NAME", "TRACE_LEVEL_VALUE"]


PACKAGE_NAME = os.path.basename(os.path.dirname(__file__))
DEFAULT_STREAM = sys.stderr
STREAMS = {"OUT": sys.stdout, "ERR": sys.stderr}

LOGGING_LEVEL = "INFO"
BASIC_LOGGING_FORMAT = "%(message)s"
DEV_LOGGING_FMT = "[%(asctime)s] {%(name)s:%(lineno)d} (%(funcName)s) [%(levelname)s] > %(message)s "
TRACE_LEVEL_VALUE = 5
TRACE_LEVEL_NAME = "TRACE"
CUSTOM_LEVELS = {TRACE_LEVEL_NAME: TRACE_LEVEL_VALUE}


SILENCE_LOGS_OPTNAME = "silent"
VERBOSITY_OPTNAME = "verbosity"
DEVMODE_OPTNAME = "logdev"
PARAM_BY_OPTNAME = {DEVMODE_OPTNAME: "devmode"}

# Translation of verbosity into logging level.
# Log message count monotonically increases in verbosity while it decreases
# in logging level, making verbosity a more intuitive specification mechanism.
_WARN_REPR = "WARN"
LEVEL_BY_VERBOSITY = ["CRITICAL", "ERROR", _WARN_REPR, "INFO", "DEBUG"]

LOGGING_CLI_OPTDATA = {
    SILENCE_LOGS_OPTNAME: {
            "action": "store_true", "help": "Silence logging"},
    VERBOSITY_OPTNAME: {
            "help": "Relative measure of interest in logs; this can be an "
                    "integer in [0, 5], or a Python builtin logging name)"},
    DEVMODE_OPTNAME: {
            "action": "store_true",
            "help": "Handle logging in development mode; perhaps among other "
                    "facets, make the format more information-rich."}
}


def add_logging_options(parser):
    """
    Augment a CLI argument parser with this package's logging options.
    
    :param argparse.ArgumentParser parser: CLI options and argument parser to
        augment with logging options.
    :return argparse.ArgumentParser: the input argument, supplemented with this
        package's logging options.
    """
    for optname, optdata in LOGGING_CLI_OPTDATA.items():
        parser.add_argument("--{}".format(optname), **optdata)
    return parser


def logger_via_cli(opts, **kwargs):
    """
    Convenience function creating a logger.
    
    This module provides the ability to augment a CLI parser with 
    logging-related options/arguments so that client applications do not need 
    intimate knowledge of the implementation. This function completes that 
    lack of burden, parsing values for the options supplied herein.
    
    :param argparse.Namespace opts: command-line options/arguments.
    :return logging.Logger: configured logger instance.
    :raise pararead.logs.AbsentOptionException: if one of the expected options
        isn't available in the given Namespace. Such a case suggests that a
        client application didn't use this module to add the expected logging
        options to a parser.

    """
    # Within the key, translate the option name if needed. If it's not
    # present within the translations mapping, use the original optname.
    # Once translation's done (if needed), parse out the
    logs_cli_args = {}
    for optname in LOGGING_CLI_OPTDATA.keys():
        # Client must add the expected options, via the API or otherwise.
        try:
            optval = getattr(opts, optname)
        except AttributeError:
            raise AbsentOptionException(optname)
        else:
            # Translate the option name if needed (i.e., for discordance
            # between the CLI version and the logger setup signature).
            logs_cli_args[PARAM_BY_OPTNAME.get(optname, optname)] = optval
    logs_cli_args.update(kwargs)
    return setup_logger(**logs_cli_args)


def setup_logger(
        stream=None, logfile=None,
        make_root=True, propagate=False, silent=False, devmode=False,
        verbosity=None, fmt=None, datefmt=None):
    """
    Establish the package-level logger.

    This is intended to be called just once per "session", with a "session"
    defined as an invocation of the main workflow, a testing session, or an
    import of the primary abstractions, e.g. in an interactive iPython session.

    :param str stream: standard stream to use as log destination. The default
        behavior is to write logs to stdout, even if null is passed here. This
        is to allow a CLI argument as input to stream parameter, where it may be
        undesirable to require specification of a default value in the client 
        application in order to prevent passing None if no CLI option value 
        is given. To disable standard stream logging, set 'silent' to True 
        or pass a path to a file to which to write logs, which gets priority 
        over a standard stream as the destination for log messages.
    :param str | FileIO[str] logfile: path to filesystem location to use as
        logs destination. if provided, this mutes standard stream logging.
    :param bool make_root: whether to use returned logger as root logger. This
        means the name will be 'root' and that messages will not propagate.
    :param bool propagate: whether to allow messages from this logger to reach
        parent logger(s).
    :param bool silent: whether to silence logging; this is only guaranteed for
        messages from this logger and for those from loggers beneath this one
        in the runtime hierarchy without no separate handling. Propagation must
        also be turned off separately--if this is not the root logger--in
        order to ensure that messages are not handled and emitted from a 
        potential parent to the logger built here.
    :param bool devmode: whether to log in development mode; possibly among
        other behavioral changes to logs handling, use a more information-rich
        message format template.
    :param int | str verbosity: alternate mode of expression for logging level
        that better accords with intuition about how to convey this. It's
        positively associated with message volume rather than negatively so, as
        logging level is. This takes precedence over 'level' if both are present.
    :param str fmt: message format/template.
    :param str datefmt: format/template for time component of a log record.
    :return logging.Logger: configured Logger instance

    """

    # Enable named ultrafine logging for debugging.
    for level_name, level_value in CUSTOM_LEVELS.items():
        logging.addLevelName(level_value, level_name)

    # Establish the logger.
    name = "" if make_root else PACKAGE_NAME
    logger = logging.getLogger(name)
    logger.handlers = []
    logger.propagate = propagate and not make_root

    # Either short-circuit with a silent logger or parse and set level.
    if silent:
        logger.addHandler(logging.NullHandler())
        return logger

    level = _level_from_verbosity(verbosity or LOGGING_LEVEL)
    logger.setLevel(level)

    # Logfile supersedes stream logging.
    if logfile:
        logfile_folder = os.path.dirname(logfile)
        if not os.path.exists(logfile_folder):
            os.makedirs(logfile_folder)
        
        # Create and add the handler, overwriting rather than appending.
        handler = logging.FileHandler(logfile, mode='w')
    
    else:
        stream = stream or DEFAULT_STREAM

        # Deal with possible argument types.
        if stream in [sys.stderr, sys.stdout]:
            stream_loc = stream
        else:
            try:
                # Assume that we have a stream-indicative text argument.
                stream_loc = STREAMS[stream.upper()]
            except (AttributeError, KeyError):
                # Fall back on default stream since
                # arguments indicate that one should be activate.
                print("Invalid stream location: {}; using {}".
                      format(stream, DEFAULT_STREAM))
                stream_loc = DEFAULT_STREAM
        
        handler = logging.StreamHandler(stream_loc)

    # Determine format.
    if not fmt:
        use_dev = devmode or isinstance(handler, logging.FileHandler) or \
                  level <= logging.DEBUG
        fmt = DEV_LOGGING_FMT if use_dev else BASIC_LOGGING_FORMAT

    handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.info("Configured logger '%s' using %s v%s",
                logger.name, PACKAGE_NAME, __version__)
    return logger


class AbsentOptionException(Exception):
    """ Exception subtype suggesting that client should add log options. """
    def __init__(self, missing_optname):
        likely_reason = "'{}' not in the parsed options; was {} used to " \
                        "add CLI logging options to an argument parser?". \
                format(missing_optname, "{}.{}".format(
                        __name__, add_logging_options.__name__))
        super(AbsentOptionException, self).__init__(likely_reason)


def _parse_level(loglevel):
    """
    Handle pitfalls of logging level specification, using fallback value.

    :param int | str loglevel : Value or name of value for logging level to use.
    :return int: Integer representation of the input value, or a default if the
        given value was unable to be parsed and interpreted as a logging level.

    """
    level = None
    try:
        level = int(loglevel)
    except (TypeError, ValueError):
        try:
            levelname = loglevel.upper()
        except AttributeError:
            pass
        else:
            try:
                level = getattr(logging, levelname)
            except AttributeError:
                try:
                    level = CUSTOM_LEVELS[levelname]
                except KeyError:
                    pass
    finally:
        if not level:
            level = LOGGING_LEVEL
            if loglevel is not None:
                print("Invalid logging level: '{}'".format(loglevel))
            print("Using {} as logging level.".format(level))
        return level


def _level_from_verbosity(verbosity):
    """
    Translation of verbosity into logging level.
    
    Log message count monotonically increases in verbosity 
    while it decreases in logging level, making verbosity 
    a more intuitive specification mechanism for users.

    :param int | str verbosity: small integral value representing a relative
        measure of interest in seeing messages about program execution,
        or the name of a Python builtin logging level
    :return int: numeric logging level in accordance with Python builtin logging

    """
    try:
        verbosity = int(verbosity)
    except:
        pass
    if isinstance(verbosity, str):
        v = verbosity.upper()
        if v.startswith(_WARN_REPR):
            v = _WARN_REPR
        if v not in LEVEL_BY_VERBOSITY:
            raise ValueError(
                "Invalid logging verbosity ('{}'); choose from: "
                "{}".format(verbosity, ", ".join(LEVEL_BY_VERBOSITY)))
        return getattr(logging, v)
    elif isinstance(verbosity, int):
        # Allow negative value to mute even ERROR level but not CRITICAL.
        # Also handle excessively high verbosity request.
        v = min(max(verbosity, 0), len(LEVEL_BY_VERBOSITY) - 1)
        return LEVEL_BY_VERBOSITY[v]
    else:
        raise TypeError("Verbosity must be string or int; got {} ({})"
                        .format(verbosity, type(verbosity)))



""" Specific exception types. """

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"



class CommandOrderException(Exception):
    """ The parallel reads processor needs certain method call sequence. """
    def __init__(self, reason=""):
        super(CommandOrderException, self).__init__(reason)



class FileTypeException(Exception):
    """ Extension not matching any of those of supported file types. """

    def __init__(self, got, known):
        """
        Declare filetype received and those supported.

        :param str got: file type, name, or path of offending file.
        :param str | Iterable[str] known: supported filetype(s).
        """
        reason = "'{}' is not among supported filetypes: {}".format(got, known)
        super(FileTypeException, self).__init__(reason)



class IllegalChunkException(Exception):
    """ Illegal reads chunk ID. """
    def __init__(self, requested, of_interest):
        reason = "Requested {} but processing was restricted to: {}".\
                format(requested, of_interest)
        super(IllegalChunkException, self).__init__(reason)



class MissingHeaderException(Exception):
    """ A reads file undergoing processing must have a header. """
    def __init__(self, filepath=""):
        reason = "No chromosomes in header; this file is " \
                 "empty or unaligned. Aligned reads are required{}".\
                 format(": '{}'".format(filepath) if filepath else ".")
        super(MissingHeaderException, self).__init__(reason)



class MissingOutputFileException(Exception):
    """
    Filepath for particular chunk output doesn't exist.
    
    Based on its internal settings, the processor's combine() step derives 
    a filepath to the output file for each reads chunk indicated by key 
    in the argument that it receives. If one of those is missing, it may 
    be considered an exceptional case.
    """
    def __init__(self, reads_chunk_key, filepath):
        reason = "Path to output file for reads chunk '{}' " \
                 "does not exist: '{}'".format(reads_chunk_key, filepath)
        super(MissingOutputFileException, self).__init__(reason)



class UnknownChromosomeException(Exception):
    """ Represent case in which data about a chromosome is not available. """
    def __init__(self, requested, known=None):
        reason = requested
        if known:
            reason += "; known: {}".format(known)
        super(UnknownChromosomeException, self).__init__(reason)



""" Parallel reads processor utilities. """

from collections import namedtuple
import itertools
import operator as op
import os
import sys
if sys.version_info < (3, 3):
    from collections import Mapping, Sequence
else:
    from collections.abc import Mapping, Sequence
from pysam import AlignmentFile, VariantFile
from .exceptions import FileTypeException, MissingHeaderException


__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


__all__ = ["create_reads_builder",
           "interleave_chromosomes_by_size",
           "make_outfile_name", "parse_bam_header",
           "partition_chunks_by_null_result",
           "pending_feature", "unbuffered_write"]


ReadsFileMaker = namedtuple("ReadsFileMaker", field_names=["ctor", "kwargs"])

# TODO: pysam docs say 'u' for uncompressed BAM.
READS_FILE_MAKER = {
    "SAM": ReadsFileMaker(AlignmentFile, {"mode": 'r'}),
    "BAM": ReadsFileMaker(AlignmentFile, {"mode": 'rb'}),
    "CRAM": ReadsFileMaker(AlignmentFile, {"mode": 'rc'}),
    "VCF": ReadsFileMaker(VariantFile, {"mode": 'r'}),
    "BCF": ReadsFileMaker(VariantFile, {"mode": 'rb'})
}


def create_reads_builder(path_reads_file):
    """
    Create the factory for a reads file.
    
    This is called when the parallel reads processor registers file(s), 
    inferring from the extension the type of file to create and supplying  
    the reads file factory, allowing additional keyword arguments to be 
    passed to the constructor before the reads file is created.
    
    :param str path_reads_file: path to file with sequencing reads data.
    :return pararead.utils.ReadsFileMaker: a namedtuple providing the proper
        pysam reads file constructor and just the most basic keyword
        argument(s), allowing the caller to add more specific keyword arguments
        before creating a reads file instance.
    :raise pararead.exceptions.FileTypeException: if the given filepath appears
        to be of an unsupported type.
    """
    _, extension = os.path.splitext(path_reads_file)
    filetype = extension[1:].upper()
    try:
        reads_file_maker = READS_FILE_MAKER[filetype]
    except KeyError:
        raise FileTypeException(got=path_reads_file,
                                known=READS_FILE_MAKER.keys())
    return reads_file_maker


def interleave_chromosomes_by_size(size_by_chromosome):
    """
    Arrange chromosome names to facilitate even binning.
    
    Intersperse/interleave chromosomes such that ones at opposite ends of 
    the spectrum of sizes are adjacent.
    
    :param Iterable[str, int] | Mapping[str, int] size_by_chromosome: pairing
        of chromosome name and size/length
    :return Iterable[str]: Names of chromosomes
    """

    if not size_by_chromosome:
        return []
    if isinstance(size_by_chromosome, Mapping):
        size_by_chromosome = size_by_chromosome.items()

    ordered_chromosomes = zip(*sorted(size_by_chromosome,
                                      key=op.itemgetter(1)))[0]
    num_chromosomes = len(ordered_chromosomes)
    meridian = int(num_chromosomes / 2)
    first_half, second_half = \
            ordered_chromosomes[:meridian], ordered_chromosomes[meridian:]

    interleaved = list(itertools.chain(*zip(first_half, second_half[::-1])))

    num_interleaved = len(interleaved)
    if num_interleaved != num_chromosomes:
        # Account for odd number of chromosomes; zip will truncate.
        assert num_interleaved == num_chromosomes - 1

        # Second half is zipped with first in reverse order, so if there's
        # an element that's gone missing, it would be the first element of
        # the second half. It will have been "flipped out" of the zip.
        # The second half is guaranteed to be no longer than the first since
        # the first is taken up to the int-truncated half of list size.
        maybe_missing_chrom = second_half[0]
        assert maybe_missing_chrom not in interleaved
        interleaved.append(maybe_missing_chrom)

    return interleaved


def make_outfile_name(readsfile_basename, processing_action, output_type):
    """
    Create a name for an output file based on action performed.

    :param str readsfile_basename: path-less and extension-less version of name
        of file of reads.
    :param str processing_action: name for the processing action being
        performed by a parallel processor of sequencing reads (i.e., a class
        derived from ParaReadProcessor).
    :param str output_type: type of output file for which name is being created.
    :return str: (Fallback) name for output file, used by the ParaReadProcessor
        constructor if a null or empty output filename is provided at creation.
    """
    return "{}_{}.{}".format(readsfile_basename,
                             processing_action, output_type)


def parse_bam_header(readsfile, chroms=None, require_aligned=False):
    """
    Get a list of chromosomes (and lengths) in this readsfile from header.

    :param pysam.libcalignmentfile.AlignmentFile readsfile: file with aligned
        sequencing reads datasets
    :param Iterable[str] chroms: chromosomes of interest; if empty/null/False,
        assume that all chromosomes with read(s) are of interest.
    :param bool require_aligned: whether to throw an exception if given
        unaligned input
    :return NoneType | Mapping[str, int]: null if no chromosomes are in the
        header (unaligned?) and non-strict (i.e., not requiring aligned input);
        otherwise, a mapping from chromosome name to length.
    :raise pararead.exceptions.MissingHeaderException: if the reads file header
        lacks chromosome names (unaligned?) and strictness is imposed (raise
        exception for this case).
    """

    try:
        all_sizes_by_chrom = {headline['SN']: headline['LN']
                              for headline in readsfile.header['SQ']}
    except KeyError:
        all_sizes_by_chrom = {}

    if not all_sizes_by_chrom:
        if require_aligned:
            raise MissingHeaderException(readsfile.filename)
        else:
            return None

    if not chroms:
        # No filtration --> retain all chromosomes.
        return all_sizes_by_chrom

    return {c: s for c, s in all_sizes_by_chrom.items() if c in set(chroms)}


def partition_chunks_by_null_result(result_by_chromosome):
    """
    Bin chromosome name by whether processing result was null.

    :param Sequence[(str, object)] | Mapping[str, object] result_by_chromosome:
        pairs of name and result of processing
    :return (list[str], list[str]): sequence of names of chromosomes for which
        result was null, and an analogous sequence for those with a null result
    """
    # Ideally, the filtration strategy would be an argument.
    # Due to cPickle's disdain for anonymous functions and general
    # inability to serialize a callable, though, the filtration
    # strategy employed is fixed to be whether the result is null.
    bad_chroms, good_chroms = [], []
    if isinstance(result_by_chromosome, Mapping):
        res_by_chr = result_by_chromosome.items()
        if not isinstance(result_by_chromosome, Sequence):
            # Non-mapping input suggests pre-sorting, so sort Mapping to
            # facilitate concordance between results for varied input types.
            # Don't destroy order of an already-ordered Mapping, though.
            res_by_chr = sorted(res_by_chr)
    else:
        res_by_chr = result_by_chromosome
    for c, r in res_by_chr:
        chroms = bad_chroms if r is None else good_chroms
        chroms.append(c)
    return bad_chroms, good_chroms


def pending_feature(not_yet_implemented):
    """
    Indicate that a callable's implementation is not complete or not stable.
    
    This simplifies designation and application of this concept, 
    and it simplifies removal once the implementation is ready for use.
    
    :param callable not_yet_implemented: the function or class not ready to be
        used, i.e. that's pending implementation.
    :return callable: object that will raise NotImplementedError if called.
    """
    def raise_error(*args, **kwargs):
        raise NotImplementedError("{} is not fully implemented".
                                  format(not_yet_implemented.__name__))
    return raise_error


def unbuffered_write(txt):
    """ Write unbuffered output by flushing after each stdout.write call. """
    sys.stdout.write(txt)
    sys.stdout.flush()



#!/usr/bin/env python
""" Install and setup the pararead package. """

from setuptools import setup
import sys

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


# Ordinary package dependencies
with open("requirements/requirements.txt", 'r') as dependencies:
    _DEPENDENCIES = []
    for line in dependencies:
        if line.startswith("#") or not line.strip():
            continue
        package = line.split("=")[0].rstrip("><")
        _DEPENDENCIES.append(package)


# Test dependencies
with open("requirements/requirements-test.txt", 'r') as test_deps_file:
    test_deps = []
    for packname in test_deps_file.readlines():
        dependency = packname.strip()
        if dependency:
            test_deps.append(dependency)


# Version info
with open("pararead/_version.py", 'r') as versionfile:
    # Assume version file like: '__version__ = "0.0.0"\n'
    _VERSION = versionfile.readline().split()[-1].strip("\"'\n")



# Handle the pypi README formatting.
try:
    import pypandoc
    _LONG_DESC = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    _LONG_DESC = open('README.md').read()


setup(
    name="pararead",
    packages=["pararead"],
    version=_VERSION,
    description="Parallel processing of sequencing reads",
    long_description=_LONG_DESC,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    keywords="bioinformatics, ngs, sequencing",
    url="https://github.com/databio/pararead",
    author=u"Nathan Sheffield, Vince Reuter",
    license="BSD2",
    install_requires=_DEPENDENCIES,
    test_suite="tests",
    tests_require=test_deps,
    setup_requires=(["pytest-runner"]
                    if {"ptr", "test", "pytest"} & set(sys.argv)
                    else []),
)



"""
Process sequencing reads in SAM or BAM format in parallel.

Do so by deriving from ParaReadProcessor, critically implementing __call__().
That defining what to do with each chromosome.
Then you simply call object.register_files(),
then object.run(), object.combine().
"""

import abc
import atexit
import itertools
import logging
import multiprocessing
import os
import shutil
import tempfile

from .exceptions import \
    CommandOrderException, IllegalChunkException, \
    MissingOutputFileException, UnknownChromosomeException
from .logs import setup_logger
from .utils import *


"""
This module-level variable acts like a global.
The ParaReadProcessor child classes will be
function classes that will be called by the
multiprocessing pool map function, which must
pickle everything in the call. To work around this,
I stick any unpicklable or large items into this
PARA_READ_FILES dict, so they aren't part of the class,
and thus will not cause map to fail.
"""
PARA_READ_FILES = {}
READS_FILE_KEY = "readsfile"
CHUNKS_PER_CORE = 5
CORES_PARAM_NAME = "cores"


_LOGGER = logging.getLogger(__name__)


class ParaReadProcessor(object):
    """
    Base class for parallel processing of sequencing reads.
    
    Implement __call__ to define work for each reads chunk, (e.g., chromosome).
    Unaligned reads are permitted, but the work then cannot rely on any sort 
    of biologically meaningful chunking of the reads unless a partition() 
    function is implemented. If unaligned reads are used and no partition() is 
    implemented, reads will be arbitrarily split into chunks.
    
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
            self, path_reads_file, cores, outfile=None, action=None,
            temp_folder_parent_path=None, limit=None, allow_unaligned=False,
            require_new_outfile=False, by_chromosome=True,
            intermediate_output_type="txt", output_type="txt",
            retain_temp=False):
        """
        :param str path_reads_file: data location (aligned BAM/SAM file).
        :param int | str cores: number of processors to use.
        :param str outfile: path to location for output file; either this
            or action is required.
        :param str action: name for what the child class is doing, used to
            derive outfile name if unspecified; if outfile is unspecified,
            this is required.
        :param str temp_folder_parent_path: temporary folder; if unspecified,
            this will match the folder containing the output file.
        :param list[str] limit: which chromosomes to process, process all by
            default.
        :param bool allow_unaligned: whether to allow unaligned reads.
        :param bool require_new_outfile: whether to raise an exception if
            output file already exists.
        :param bool by_chromosome: whether to chunk reads on a per-chromosome
            basis, implicitly imposing requirement for aligned reads.
        :param str intermediate_output_type: type of output file generated for
            each chunk of reads processed.
        :param str output_type: type of final output file generated; this is
            used by both intermediate files that are created and by the combine()
            step that creates final output.
        :raise ValueError: if given neither `outfile` path nor `action` action
            name, or if output file already exists and a new one is required.
        """

        # Establish root logger only if client application hasn't done so.
        # That is, create a root logger with a handler if one doesn't exist.
        if not logging.getLogger().handlers:
            global _LOGGER
            _LOGGER = setup_logger(make_root=True, propagate=False)

        # Initial path setup and filetype handling.
        name_reads_file = os.path.basename(path_reads_file)
        readsfile_basename, _ = os.path.splitext(name_reads_file)
        self.path_reads_file = path_reads_file

        # Use given output path or derive it from processing behavior.
        if outfile:
            self.outfile = outfile
        elif action:
            self.outfile = make_outfile_name(
                    readsfile_basename, action, output_type)
        else:
            raise ValueError("Either path to output file or "
                             "name of processing action is required.")

        # Perform check after establishing the setting so that it works
        # regardless of whether the path was explicit or inferred.
        if os.path.exists(self.outfile):
            if require_new_outfile:
                raise ValueError("Output file already exists: '{}'".
                                 format(self.outfile))
            else:
                _LOGGER.warning(
                        "Output file already exists and "
                        "will be overwritten: '{}'".format(self.outfile))

        # Create temp folder that's deleted upon exit.
        if not temp_folder_parent_path:
            temp_folder_parent_path = os.path.dirname(self.outfile)

        # Handling of temporary folder.
        prefix = "tmp_{}_".format(readsfile_basename)
        if action:
            prefix += "{}_".format(action)
        tempfolder = tempfile.mkdtemp(
                prefix=prefix, dir=temp_folder_parent_path)
        self.temp_folder = tempfolder

        # Add a couple lines so that tests can execute quietly.
        # Tests handle cleanup separately, so if this existence
        # check is omitted, exceptions can be squelched, but their
        # messages can still appear due to rmtree()'s use of sys.exc_info().
        def clean():
            if os.path.exists(tempfolder):
                shutil.rmtree(tempfolder)
        
        if not retain_temp:
            atexit.register(clean)

        # Behavior/execution parameters.
        self.cores = int(cores)
        self.limit = limit
        self.require_aligned = by_chromosome or not allow_unaligned
        self.intermediate_output_type = intermediate_output_type
        self.by_chromosome = by_chromosome
        self._size_by_chromosome = None

    @abc.abstractmethod
    def __call__(self, chunk_id, reads_chunk):
        """
        Perform 'chunk'-wise processing implemented in the subclass.

        A concrete implementation operates on a chunk of sequencing reads. 
        By default, a 'chunk' simply consists of contiguous reads from the 
        input file, but a subclass that overrides the partitioning mechanism 
        can change that. In that case, the provided 'chunk_id' will likely 
        be meaningful, but otherwise it can be ignored. If processing fails 
        for the given chunk, the concrete implementation should communicate 
        this by returning None.

        :param int | str chunk_id: reads chunk identifier; this is passed in
            case it's of use in the client implementation, but for many use
            cases it can be ignored.
        :param Iterable reads_chunk: chunk of sequencing reads to process,
            likely pysam.AlignedSegment
        :return NoneType | object: null if the chunk's processing failed,
            otherwise a non-null result.
        """
        pass

    @property
    def files(self):
        """
        Refer to the pararead files mapping.
        
        :return Mapping[str, object]: pararead files mapping.
        """
        return PARA_READ_FILES

    @property
    def readsfile(self):
        """
        :return pysam.AlignmentFile | pysam.VariantFile: instance of the reads
            file abstraction appropriate for the given type of input data
            (e.g., BAM or VCF).
        :raise pararead.exceptions.CommandOrderException: if a command
            prerequisite for a parallel reads processor operation has not yet
            been performed.
        """
        return self.fetch_file(READS_FILE_KEY)


    @staticmethod
    def empty_action(read_chunk_key=None):
        """
        Action to take when processing an empty reads chunk.

        :param str read_chunk_key: key for the empty chunk of reads.
        """
        if read_chunk_key:
            _LOGGER.debug("Empty read chunk: {}".format(read_chunk_key))

    def fetch_file(self, file_key):
        """
        Retrieve one of the files registered with pararead.

        :param str file_key: which file to fetch
        :return object: likely pysam.AlignmentFile -- file ADT instance
            associated with the requested key.
        :raise pararead.exceptions.CommandOrderException: if the indicated file
            hasn't been registered.
        """
        try:
            return self.files[file_key]
        except KeyError:
            raise CommandOrderException(
                "No {} established; has {} been called?".format(
                    READS_FILE_KEY, ParaReadProcessor.register_files.__name__))

    def check_command(self, cmd):
        """
        Determine whether it appears that a command may be run.

        :param str cmd: command to check for runnability
        :return OSError: if it's possible to verify that running given command
            would fail
        """
        if not self.is_command_callable(cmd):
            raise OSError("{} is not callable".format(cmd))

    def is_command_callable(self, command, name=""):
        """
        Check if command can be called.

        :param str command: actual command to call
        :param str name: nickname/alias by which to reference the command
        :return bool: whether given command's call succeeded
        """
        name = name or command
        # Use `command` to see if command is callable, store exit code
        code = os.system(
            "command -v {0} >/dev/null 2>&1 || {{ exit 1; }}".format(command))
        if code != 0:
            alias_value = " ('{}') ".format(name) if name else " "
            _LOGGER.debug("Command '{0}' is not callable: {1}".
                          format(alias_value, command))
        return not bool(code)

    def get_chrom_size(self, chrom):
        """
        Determine the size of the given chromosome.
        
        :param str chrom: name of chromosome of interest.
        :return int: size of chromosome of interest.
        :raise pararead.exceptions.CommandOrderException: if there's no
            chromosome sizes map yet.
        :raise pararead.exceptions.UnknownChromosomeException: if requested
            chromosome is not in the sizes map.
        """
        if not self._size_by_chromosome:
            raise CommandOrderException(
                    "No size-by-chromosome mapping; "
                    "has a reads file been registered?")
        try:
            return self._size_by_chromosome[chrom]
        except KeyError:
            raise UnknownChromosomeException(
                    chrom, known=self._size_by_chromosome.keys())


    def register_files(self, **file_builder_kwargs):
        """
        Add to module map any large/unpicklable variables required by __call__.

        :raise pararead.exceptions.FileTypeException: if path to the reads file
            given doesn't appear to match one of the supported file types.
        """

        _LOGGER.info("Registering input file: '%s'", self.path_reads_file)
        reads_file_maker = create_reads_builder(self.path_reads_file)

        # Here, check_sq is necessary so that ParaRead can process
        # unaligned files, which is occasionally desirable.
        builder = reads_file_maker.ctor
        kwargs = file_builder_kwargs

        # Builder has minimal requirements that must be met.
        # Thus, those take precedence over user provisions.
        kwargs.update(reads_file_maker.kwargs)

        if not self.require_aligned:
            kwargs['check_sq'] = False

        readsfile = builder(self.path_reads_file, **kwargs)
        PARA_READ_FILES[READS_FILE_KEY] = readsfile

        # Cache mapping from chromosome name to size for easy access.
        self._size_by_chromosome = parse_bam_header(
                readsfile, require_aligned=self.require_aligned)

        def ensure_closed():
            if readsfile.is_open:
                readsfile.close()
        atexit.register(ensure_closed)

    def run(self, chunksize=None, interleave_chunk_sizes=False):
        """
        Do the processing defined partitioned across each unit (chromosome).

        :param int chunksize: number of reads per processing chunk; if
            unspecified, the default heuristic of size s.t. each core gets ~ 4
            chunks.
        :param bool interleave_chunk_sizes: whether to interleave reads chunk
            sizes. If off (default), just use the distribution that Python
            determines.
        :return Iterable[str]: names of chromosomes for which result is non-null.
        :raise pararead.exception.MissingHeaderException: if attempting to run
            with an unaligned reads file in the context of an aligned file
            requirement.
        """
        # Because this class is a function class (implements __call__), I can
        # call "self()" as a function, which is what runs the match function
        # on a single chromosome. By mapping self() across multiple chroms,
        # I get the parallel version.

        try:
            readsfile = PARA_READ_FILES[READS_FILE_KEY]
        except KeyError:
            _LOGGER.error(
                    "No '{}' has been established; call 'register_files' "
                    "before 'run'".format(READS_FILE_KEY))
            raise

        if not self.by_chromosome:
            read_chunk_keys = self.chunk_reads(readsfile, chunksize=chunksize)
        else:
            size_by_chromosome = parse_bam_header(
                readsfile=readsfile, chroms=self.limit,
                require_aligned=self.require_aligned)
            if self.cores == 1:
                # TODO: handle case (unaligned input) of null return.
                # TODO: pysam's fetch() may make this OK but not distribute.
                read_chunk_keys = size_by_chromosome.keys()
            else:
                if size_by_chromosome is None:
                    # Unaligned files lack any chrom header lines.
                    # If so, we'll get back a null to disambiguate
                    # empty filter case if permitting unaligned
                    # input. If requiring aligned input, the call
                    # will have already generated an exception to that effect.
                    _LOGGER.warning(
                            "Failed attempt to parse chromosomes as read "
                            "chunk keys; arbitrarily chunking reads instead.")
                    read_chunk_keys = self.chunk_reads(
                            readsfile, chunksize=chunksize)
                else:
                    if interleave_chunk_sizes:
                        # Interleave chromosomes by size so that if tasks are
                        # pre-allocated to workers, we'll get about even bins.
                        read_chunk_keys = interleave_chromosomes_by_size(
                                size_by_chromosome.items())
                    else:
                        read_chunk_keys = size_by_chromosome.keys()

        _LOGGER.info("Temporary files will be stored in: '{}'".
                     format(self.temp_folder))
        _LOGGER.info("Processing with {} cores...".format(self.cores))

        # Some implementors may have a strand mode attribute.
        # If so, log it here to avoid duplicate messaging, as it
        # will remain constant across processed chunks (chromosomes).
        try:
            _LOGGER.info("STRAND MODE: {}".format(self.use_strand))
        except AttributeError:
            pass

        # TODO: handle non-chromosome-based case.
        idxstats = readsfile.get_index_statistics()
        reads_by_chrom = {istat.contig: istat.total for istat in idxstats}
        empties, nonempties = [], []
        for c in read_chunk_keys:
            target = empties if 0 == reads_by_chrom[c] else nonempties
            target.append(c)

        # Maps for order preservation. This permits arbitrary result return,
        # i.e. something other than the chunk key itself, when the process
        # completes. It would be a bit simpler to filter on the results
        # directly, but clients may want to implement a processor that
        # has a result with meaning beyond a signal/flag that it succeeded
        # for a particular chunk ID. That is, it may produce a result with
        # downstream meaning, and not be used simply for effect on disk.
        if self.cores == 1:
            results = map(self, nonempties)
        else:
            workers = multiprocessing.Pool(self.cores)
            # The typical call to map fails to acknowledge KeyboardInterrupts.
            # This fix helps: http://stackoverflow.com/a/1408476/946721

            results = workers.map_async(self, nonempties).get(9999999)

        # TODO: note the dependence on order here.
        result_by_chunk = [(c, self.empty_action(c)) for c in empties] + \
                           list(zip(nonempties, results))
        bad_chunks, good_chunks = \
                partition_chunks_by_null_result(result_by_chunk)

        if bad_chunks:
            _LOGGER.info("Discarding {} chunk(s) of reads: {}".
                         format(len(bad_chunks), bad_chunks))
            _LOGGER.info("Keeping {} chunk(s) of reads: {}".
                         format(len(good_chunks), good_chunks))
        else:
            _LOGGER.info("Using all reads")

        return good_chunks

    def fetch_chunk(self, chromosome):
        """
        Pull a chunk of sequencing reads from a file.
        
        :param str chromosome: identifier for chunk of reads to select.
        :return Iterable[pysam.AlignedSegment]: collection of aligned reads
        """
        if not self.by_chromosome:
            raise NotImplementedError(
                    "Provide a fetch_chunk implementation "
                    "if not partitioning reads by chromosome.")
        readsfile = PARA_READ_FILES[READS_FILE_KEY]
        return readsfile.fetch(chromosome, multiple_iterators=True)

    def combine(self, good_chromosomes, strict=False, chrom_sep=None):
        """
        Aggregate output from independent read chunks into single output file.
        
        :param Iterable[str] good_chromosomes: identifier (e.g., chromosome)
            for each chunk of reads processed.
        :param bool strict: whether to throw an exception upon encountering a
            missing file. If not, simply log a warning message and continue the
            aggregation process that's underway, working with what is available.
        :param str chrom_sep: delimiter between output from each chromosome.
        :return Iterable[str]: path to each file successfully combined.
        :raise pararead.exceptions.MissingOutputFileException: if executing in
            strict mode, and there's a reads chunk key for which the derived
            filepath does not exist.
        :raise pararead.exceptions.IllegalChunkException: if a chunk of reads
            outside of those declared to be of interest is requested to
            participate in the combination.
        """

        if not good_chromosomes:
            _LOGGER.warning("No successful chromosomes, so no combining.")
            return

        # Check that the combination request accords with the chunks
        # declared to be of interest
        if self.limit:
            missing_chunks = set(good_chromosomes) - set(self.limit)
            if missing_chunks:
                raise IllegalChunkException(
                        requested=missing_chunks, of_interest=self.limit)

        _LOGGER.info("Merging {} files into output file: '{}'".
                     format(len(good_chromosomes), self.outfile))

        # Track what we actually combine (particularly if non-strict
        # with respect to chunk(s) for which output file is missing.
        paths_combined_files = []

        if len(good_chromosomes) == 1:
            _LOGGER.debug("Just one good chromosome; ignoring delimiter.")
            chrom_sep = None

        with open(self.outfile, 'w') as outfile:
            for chrom in good_chromosomes:
                reads_chunk_output = self._tempf(chrom)

                # Handle case in which chunk's output is missing.
                if not os.path.exists(reads_chunk_output):
                    if strict:
                        raise MissingOutputFileException(
                                reads_chunk_key=chrom,
                                filepath=reads_chunk_output)
                    else:
                        _LOGGER.warning(
                                "Missing output file for reads chunk '%s', "
                                "skipping: '%s'", chrom, reads_chunk_output)
                        continue

                # Append lines from this chunk's output.
                with open(reads_chunk_output, 'r') as tmpf:
                    for line in tmpf:
                        outfile.write(line)
                if chrom_sep:
                    outfile.write(chrom_sep)
                paths_combined_files.append(reads_chunk_output)

        return paths_combined_files

    @pending_feature
    def chunk_reads(self, readsfile, chunksize=None):
        """
        Partition sequencing reads into equally-sized 'chunks' for 
        parallel processing. This treats all reads equally, in that 
        they are grouped by contiguous occurrence within the given 
        input file. This facilitates processing of unaligned reads, 
        but it means that reads from the same chromosome will not be 
        processed together. This can be overridden if that's desired.

        :param Iterable readsfile: collection of reads to chunk, likely
            pysam.AlignmentFile pysam.VariantFile; reads to split into chunks.
        :param int chunksize: number of units (i.e., reads) per processing chunk.
            If unspecified, this is derived using the instance's 
            cores count and chunks-per-core parameter, along with 
            a count of the number of units (reads). Note that if 
            this is unspecified, there will be some additional time 
            used to count the reads to derive chunk size.
        :return Iterable[(int, itertools.groupby)]: pairs of chunk key/ID and
            chunk reads chunk itself.
        """
        if not chunksize or chunksize < 1:
            _, reads_clone = itertools.tee(readsfile)
            num_chunks = self.cores * CHUNKS_PER_CORE

            # Count the reads.
            _LOGGER.info("Deriving chunk size for %d chunks: "
                         "%d cores x %d chunks/core",
                         num_chunks, self.cores, CHUNKS_PER_CORE)
            _, reads_clone = itertools.tee(readsfile)
            num_reads = sum(1 for _ in reads_clone)
            _LOGGER.info("Reads count: %d", num_reads)

            chunksize = int(num_reads / num_chunks)

        return itertools.groupby(
            enumerate(readsfile), key=lambda ipair: int(ipair[0] / chunksize))

    def _tempf(self, chrom):
        """
        Derive name for temporary file from chromosome name.

        :param str chrom: name of chromosome (single processing partition).
        :return str: name for tempfile corresponding to given unit name.

        """
        return os.path.join(
                self.temp_folder,
                "{}.{}".format(chrom or "ALL", self.intermediate_output_type))
